/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    int JoinBufferPoolSize;                         //B - 2
    RmRecord* JoinBuffer;
    bool left_over = false;
    int left_size_this_pass = 0;
    int left_idx = 0;
    bool isend = false;
    std::vector<ColMeta> lhs_cols;
    std::vector<ColMeta> rhs_cols;

   public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, 
                            std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);
        unsigned long left_size = 0;
        //for (left_->beginTuple(); !left_->is_end(); left_->nextTuple()) left_size++;
        //JoinBufferPoolSize = std::min(left_size, MAX_MEMORY_BYTE / (left_->tupleLen() + 5) - 2);
        JoinBufferPoolSize = 262144 * 2 / (left_->tupleLen() + 5);
        if (JoinBufferPoolSize < 3)
        {
            JoinBufferPoolSize = 134217728 / (left_->tupleLen() + 5);
            if (JoinBufferPoolSize < 3) JoinBufferPoolSize = 3;
        }
        JoinBuffer = new RmRecord[JoinBufferPoolSize];
        for (int i = 0; i < JoinBufferPoolSize; i++)
            JoinBuffer[i] = RmRecord(left_->tupleLen());
        lhs_cols.resize(fed_conds_.size());
        rhs_cols.resize(fed_conds_.size());
    }

    void getLeft()
    {
        left_size_this_pass = 0;
        left_idx = 0;
        for (int i = 0; i < JoinBufferPoolSize; i++)
        {
            JoinBuffer[i].SetData(left_->Next().get()->data), left_size_this_pass++;
            left_->nextTuple();

            if (left_->is_end())
            {
                left_over = true;
                break;
            }
        }
    }

    void beginTuple() override {
        left_->beginTuple();
        if (left_->is_end())
        {
            isend = true;
            delete[] JoinBuffer;
            return;
        }
        right_->beginTuple();
        if (right_->is_end())
        {
            isend = true;
            delete[] JoinBuffer;
            return;
        }
        int idx = 0;
        for (auto cond : fed_conds_)
        {
            auto l = left_->get_col_offset(cond.lhs_col);
            auto r = right_->get_col_offset(cond.rhs_col);
            lhs_cols[idx] = l;
            rhs_cols[idx] = r;
            idx ++;
        }
        //BNLJ
        getLeft();
        bool flag = 0;
        while (left_idx < left_size_this_pass || !left_over || !right_->is_end())
        {
            while (!right_->is_end())
            {
                if (left_idx == left_size_this_pass) left_idx = 0;
                for (; left_idx < left_size_this_pass; left_idx++)
                    if (conds_satisfied(fed_conds_, &JoinBuffer[left_idx], right_->Next().get()))
                    {
                        flag = 1;
                        break;
                    }
                if (flag) break;
                right_->nextTuple();
            }
            if (flag) break;
            right_->beginTuple();
            if (!left_over) getLeft(); else {isend = true; delete[] JoinBuffer; return;}
        }
        if (!flag) { isend = true; delete[] JoinBuffer; return;}
    }

    void nextTuple() override {
        bool flag = 0;
        left_idx ++;
        if (left_idx == left_size_this_pass && !right_->is_end()) right_->nextTuple();
        while (left_idx < left_size_this_pass || !left_over || !right_->is_end())
        {
            while (!right_->is_end())
            {
                if (left_idx == left_size_this_pass) left_idx = 0;
                for (; left_idx < left_size_this_pass; left_idx++)
                    if (conds_satisfied(fed_conds_, &JoinBuffer[left_idx], right_->Next().get()))
                    {
                        flag = 1;
                        break;
                    }
                if (flag) break;
                right_->nextTuple();
            }
            if (flag) break;
            right_->beginTuple();
            if (!left_over) getLeft(); else { isend = true; delete[] JoinBuffer; return;}
        }
        if (!flag) { isend = true; delete[] JoinBuffer; return;}
    }

    std::unique_ptr<RmRecord> Next() override {
        auto record = std::make_unique<RmRecord>(len_);
        memcpy(record->data, JoinBuffer[left_idx].data, left_->tupleLen());
        memcpy(record->data + left_->tupleLen(), right_->Next()->data, right_->tupleLen());
        return record;
    }

    Rid &rid() override { return _abstract_rid; }
    bool cond_satisfied(const Condition &cond, const RmRecord *left, const RmRecord *right, int idx)
    {
        bool result = true;
        auto lhs_col = lhs_cols[idx];
        char *lhs = left->data + lhs_col.offset;

        auto rhs_col = rhs_cols[idx];
        char *rhs = right->data + rhs_col.offset;

        int cmp;                                            //比较结果
        if (lhs_col.type == TYPE_INT)
        {
            int* lhs_value = (int *)lhs;
            int rhs_value;
                    if (rhs_col.type== TYPE_INT)
                        rhs_value = *(long long*)rhs;
                    else
                        rhs_value = *(int*)rhs;
            cmp = (*lhs_value < rhs_value) ? -1 : ((*lhs_value > rhs_value) ? 1 : 0);
        }
        else if (lhs_col.type == TYPE_INT)
                {
                    long long* lhs_value = (long long *)lhs;
                    long long* rhs_value = (long long *)rhs;
                    cmp = (*lhs_value < *rhs_value) ? -1 : ((*lhs_value > *rhs_value) ? 1 : 0);
                }
        else if (lhs_col.type == TYPE_FLOAT)
        {
            double* lhs_value = (double *)lhs;
            double* rhs_value = (double *)rhs;
            cmp = (*lhs_value < *rhs_value) ? -1 : ((*lhs_value > *rhs_value) ? 1 : 0);
        }
        else if (lhs_col.type == TYPE_STRING)
        {
            std::tm* lhs_value = (std::tm *)lhs;
            std::tm* rhs_value = (std::tm *)rhs;
            std::tm ll = *lhs_value, rr = *rhs_value;
            if (ll.tm_year == rr.tm_year && ll.tm_mon == rr.tm_mon && ll.tm_mday == rr.tm_mday && ll.tm_hour == rr.tm_hour && ll.tm_min == rr.tm_min && ll.tm_sec == rr.tm_sec)
                cmp = 0;
            else cmp = 1;
        }
        else
        {
            cmp = memcmp(lhs, rhs, lhs_col.len);
        }
        if (cond.op == OP_EQ) {
            result &= (cmp == 0);
        } else if (cond.op == OP_NE) {
            result &= (cmp != 0);
        } else if (cond.op == OP_LT) {
            result &= (cmp < 0);
        } else if (cond.op == OP_GT) {
            result &= (cmp > 0);
        } else if (cond.op == OP_LE) {
            result &= (cmp <= 0);
        } else if (cond.op == OP_GE) {
            result &= (cmp >= 0);
        }
        return result;
    }
    bool conds_satisfied(const std::vector<Condition> &conds, const RmRecord *left, const RmRecord *right)
    {
        int idx = 0;
        for (auto cond : conds)
        {
            if (cond_satisfied(cond, left, right, idx) == 0) return 0;
            idx++;
        }
        return 1;
    }
    ColMeta get_col_offset(const TabCol &target) override {
        auto left_pos = left_->get_col_offset(target);
        if(left_pos.len == -1){
            auto right_pos = right_->get_col_offset(target);
            ColMeta col = {.tab_name = right_pos.tab_name,
               .name = right_pos.name,
               .type = right_pos.type,
               .len = right_pos.len,
               .offset = right_pos.offset + (int)left_->tupleLen(),
               .index = right_pos.index};
            return col;
        }
        return left_pos;
    }
    std::string getType() override { return "Join"; }
    size_t tupleLen() const override { return len_; }
    const std::vector<ColMeta> &cols() const override { return cols_; }
    bool is_end() const override { return isend; }
};