/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;              // 表的名称
    std::vector<Condition> conds_;      // scan的条件
    RmFileHandle *fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // scan后生成的记录的字段
    size_t len_;                        // scan后生成的每条记录的长度
    std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同

    Rid rid_;
    std::unique_ptr<RecScan> scan_;     // table_iterator

    SmManager *sm_manager_;
    int last_fetch = -1;
    int cnt = 0;
    std::vector<ColMeta> lhs_cols;
    std::vector<ColMeta> rhs_cols;

   public:
    SeqScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = std::move(tab_name);
        conds_ = std::move(conds);
        TabMeta &tab = sm_manager_->db_.get_table(tab_name_);
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab.cols;
        len_ = cols_.back().offset + cols_.back().len;

        context_ = context;

        fed_conds_ = conds_;
        lhs_cols.resize(fed_conds_.size());
        rhs_cols.resize(fed_conds_.size());
        int idx = 0;
        for (auto &cond : fed_conds_)
        {
            auto lhs_col = get_col(cols_, cond.lhs_col);
            lhs_cols[idx] = *lhs_col;
            if (cond.is_rhs_val)
            {
                rhs_cols[idx] = *lhs_col;
            }
            else
            {
                auto rhs_col = get_col(cols_, cond.rhs_col);    //取出对应的列
                rhs_cols[idx] = *rhs_col;
            }
            idx++;
        }
        
        //sm_manager_->get_bpm()->pre_fetch(PageId{fh_->GetFd(), -1}, 1, fh_->get_file_hdr().num_pages);
    }

    void beginTuple() override {
        scan_ = std::make_unique<RmScan>(fh_);
        findAvailable();
        //prefetch();
    }

    void nextTuple() override {
        scan_->next();
        findAvailable();
        //prefetch();
    }

    void findAvailable()
    {
        int tot = 0;
        while (!scan_->is_end()) 
        {
            //prefetch();
            rid_ = scan_->rid();
            auto rec = fh_->get_record(rid_, context_);

            bool result = true;
            int idx = 0;
            for (auto &cond : fed_conds_)
            {
                auto lhs_col = lhs_cols[idx];
                char *lhs = rec->data + lhs_col.offset;
                char *rhs;
                ColType rhs_type;
                if (cond.is_rhs_val)
                {
                    rhs_type = cond.rhs_val.type;
                    rhs = cond.rhs_val.raw->data;
                }
                else
                {
                    auto rhs_col = rhs_cols[idx];
                    rhs_type = rhs_col.type;
                    rhs = rec->data + rhs_col.offset;
                }
                if (lhs_col.type != rhs_type && (lhs_col.type != TYPE_INT))
                {
                    result = false;
                    break;
                }

                int cmp;                                            //比较结果
                if (lhs_col.type == TYPE_INT)
                {
                    int lhs_value = *(int *)lhs;
                    int rhs_value;
                    if (rhs_type == TYPE_INT)
                        rhs_value = *(long long *)rhs;
                    else
                        rhs_value = *(int *)rhs;
                    cmp = (lhs_value < rhs_value) ? -1 : ((lhs_value > rhs_value) ? 1 : 0);
                }
                else if (lhs_col.type == TYPE_INT)
                {
                    long long* lhs_value = (long long *)lhs;
                    long long* rhs_value = (long long *)rhs;
                    cmp = (*lhs_value < *rhs_value) ? -1 : ((*lhs_value > *rhs_value) ? 1 : 0);
                }
                else if (lhs_col.type == TYPE_FLOAT)
                {
                    double* lhs_value = (double *)lhs;
                    double* rhs_value = (double *)rhs;
                    cmp = (*lhs_value < *rhs_value) ? -1 : ((*lhs_value > *rhs_value) ? 1 : 0);
                }
                else if (lhs_col.type == TYPE_INT)
                {
                    std::tm* lhs_value = (std::tm *)lhs;
                    std::tm* rhs_value = (std::tm *)rhs;
                    std::tm ll = *lhs_value, rr = *rhs_value;
                    if (ll.tm_year == rr.tm_year && ll.tm_mon == rr.tm_mon && ll.tm_mday == rr.tm_mday && ll.tm_hour == rr.tm_hour && ll.tm_min == rr.tm_min && ll.tm_sec == rr.tm_sec)
                        cmp = 0;
                    else
                    {
                        if (ll.tm_year != rr.tm_year)
                            cmp = (ll.tm_year < rr.tm_year) ? -1 : 1;
                        else if (ll.tm_mon != rr.tm_mon)
                            cmp = (ll.tm_mon < rr.tm_mon) ? -1 : 1;
                        else if (ll.tm_mday != rr.tm_mday)
                            cmp = (ll.tm_mday < rr.tm_mday) ? -1 : 1;
                        else if (ll.tm_hour != rr.tm_hour)
                            cmp = (ll.tm_hour < rr.tm_hour) ? -1 : 1;
                        else if (ll.tm_min != rr.tm_min)
                            cmp = (ll.tm_min < rr.tm_min) ? -1 : 1;
                        else if (ll.tm_sec != rr.tm_sec)
                            cmp = (ll.tm_sec < rr.tm_sec) ? -1 : 1;
                    }
                }
                else
                {
                    cmp = memcmp(lhs, rhs, lhs_col.len);
                }
                if (cond.op == OP_EQ) {
                    result &= (cmp == 0);
                } else if (cond.op == OP_NE) {
                    result &= (cmp != 0);
                } else if (cond.op == OP_LT) {
                    result &= (cmp < 0);
                } else if (cond.op == OP_GT) {
                    result &= (cmp > 0);
                } else if (cond.op == OP_LE) {
                    result &= (cmp <= 0);
                } else if (cond.op == OP_GE) {
                    result &= (cmp >= 0);
                }
                if (result == false) break;
                idx++;
            }
            if (result == true)
            {
                break;
            }
            else scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        auto ret = fh_->get_record(rid_, context_);
        return ret;
    }
    void prefetch()
    {
        if (rid_.page_no != last_fetch)
        {
            if (rid_.page_no == last_fetch + 1)
            {
                last_fetch = rid_.page_no;
                cnt++;
                if (cnt % 2 == 0)
                {
                    sm_manager_->get_bpm()->pre_fetch(PageId{fh_->GetFd(), last_fetch}, 2, fh_->get_file_hdr().num_pages);
                }
            }
            else
            {
                last_fetch = rid_.page_no;
                cnt = 0;
            }
        }
    }
    Rid &rid() override { return rid_; }
    bool is_end() const override { return scan_->is_end(); }
    size_t tupleLen() const override { return len_; }
    const std::vector<ColMeta> &cols() const override { return cols_; }

    ColMeta get_col_offset(const TabCol &target) override {
        auto pos = std::find_if(cols_.begin(), cols_.end(), [&](const ColMeta &col) {
            return col.tab_name == target.tab_name && col.name == target.col_name;
        });
        if (pos == cols_.end()) {
            throw ColumnNotFoundError(target.tab_name + '.' + target.col_name);
        }
        return *pos;
    }
    std::string getType() override { return "SeqScan"; }
};