/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle *fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager *sm_manager_;

   public:
    UpdateExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
        if (context_ != nullptr && !rids_.empty())
        {
            context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd());
        }
    }
    std::unique_ptr<RmRecord> Next() override {
        for (auto &rid : rids_) {
            if (context_ != nullptr)
            {
                context_->lock_mgr_->lock_exclusive_on_record(context_->txn_, rid, sm_manager_->fhs_.at(tab_name_)->GetFd());
            }
            auto rec = fh_->get_record(rid, context_);
            RmRecord old_record(rec->size, rec->data);

            for (auto &set_clause : set_clauses_) {
                auto lhs_col = tab_.get_col(set_clause.lhs.col_name);
                memcpy(rec->data + lhs_col->offset, set_clause.rhs.raw->data, lhs_col->len);
            }
            
            bool index_chongfu = 0;
            for (size_t i = 0; i < tab_.indexes.size(); ++i) 
            {
                auto& index = tab_.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                char *new_key = new char[index.col_tot_len];
                char *old_key = new char[index.col_tot_len];
                int offset = 0;
                for (auto &col : index.cols)
                {
                    memcpy(new_key + offset, rec->data + col.offset, col.len);
                    memcpy(old_key + offset, old_record.data + col.offset, col.len);
                    offset += col.len;
                }
                /*Rid* ridd_ = new Rid();
                bool exist_ = ih->exist_key(new_key, ridd_);
                if (exist_ && *ridd_ != rid) index_chongfu = 1;*/
                ih->delete_entry(old_key, context_->txn_);
                if (ih->insert_entry(new_key, rid, context_->txn_) == -1)
                {
                    ih->insert_entry(old_key, rid, context_->txn_);
                    for (int j = i - 1; j >= 0; j--)
                    {
                        auto& index = tab_.indexes[j];
                        auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                        char *new_key1 = new char[index.col_tot_len];
                        char *old_key1 = new char[index.col_tot_len];
                        int offset = 0;
                        for (auto &col : index.cols)
                        {
                            memcpy(new_key1 + offset, rec->data + col.offset, col.len);
                            memcpy(old_key1 + offset, old_record.data + col.offset, col.len);
                            offset += col.len;
                        }
                        ih->delete_entry(new_key1, context_->txn_), 
                        ih->insert_entry(old_key1, rid, context_->txn_);
                        delete[] new_key1;
                        delete[] old_key1;
                    }
                    delete[] new_key;
                    delete[] old_key;
                    throw IndexExistsError("修改后索引列不唯一", {});
                }
                delete[] new_key;
                delete[] old_key;
            }
            if (!index_chongfu)
            {
                RmRecord new_record(rec->size, rec->data);
                //add in 11
                // UpdateLogRecord* update_log = new UpdateLogRecord(context_->txn_->get_transaction_id(), old_record, new_record, rid, tab_name_);
                // update_log->prev_lsn_ = context_->txn_->get_prev_lsn();
                // context_->log_mgr_->add_log_to_buffer(update_log);
                // context_->txn_->set_prev_lsn(update_log->lsn_);
                // delete update_log;
                WriteRecord * write_record = new WriteRecord(WType::UPDATE_TUPLE, tab_name_, rid, old_record);
                context_->txn_->append_write_record(write_record);

                fh_->update_record(rid, rec->data, context_);
                
            }
            else throw IndexExistsError("修改后索引列不唯一", {});
        }
        return nullptr;

    }

    Rid &rid() override { return _abstract_rid; }
};