/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"
#include "record/rm_file_handle.h"
#include "system/sm_manager.h"

std::unordered_map<txn_id_t, Transaction *> TransactionManager::txn_map = {};

/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction * TransactionManager::begin(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 判断传入事务参数是否为空指针
    // 2. 如果为空指针，创建新事务
    // 3. 把开始事务加入到全局事务表中
    // 4. 返回当前事务指针
    
    std::unique_lock<std::mutex> lock(latch_);
    if(txn == nullptr)
        txn = new Transaction(next_txn_id_++);
    txn->set_start_ts(next_timestamp_++);
    txn_map[txn->get_transaction_id()] = txn;
    lock.unlock();
    BeginLogRecord* begin_log = new BeginLogRecord(txn->get_transaction_id());
    begin_log->prev_lsn_ = txn->get_prev_lsn();
    log_manager->add_log_to_buffer(begin_log);
    txn->set_prev_lsn(begin_log->lsn_);
    delete begin_log;
    return txn;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 如果存在未提交的写操作，提交所有的写操作
    // 2. 释放所有锁
    // 3. 释放事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    auto lock_set = txn->get_lock_set();
    for(auto iter = lock_set->begin(); iter != lock_set->end(); iter ++)
        lock_manager_->unlock(txn, *iter);
    txn->get_lock_set()->clear();
    txn->get_write_set()->clear();
    // CommitLogRecord* commit_log = new CommitLogRecord(txn->get_transaction_id());
    // commit_log->prev_lsn_ = txn->get_prev_lsn();
    // log_manager->add_log_to_buffer(commit_log);
    // txn->set_prev_lsn(commit_log->lsn_);
    // log_manager->flush_log_to_disk();
    // txn->set_state(TransactionState::COMMITTED);
    // delete commit_log;
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction * txn, LogManager *log_manager) {
    // Todo:
    // 1. 回滚所有写操作
    // 2. 释放所有锁
    // 3. 清空事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    auto write_set = txn->get_write_set();
    Context *context = new Context(lock_manager_, log_manager, txn);
    while (!write_set->empty())
    {
        auto &item = write_set->back();                     //一定是从后往前回滚
        RmFileHandle *file_handle = sm_manager_->fhs_.at(item->GetTableName()).get();
        if (item->GetWriteType() == WType::INSERT_TUPLE)
        {
            auto record = file_handle->get_record(item->GetRid(), context);

            file_handle->delete_record(item->GetRid(), context);
            //add in 11
            file_handle->fetch_page_handle(item->GetRid().page_no).page->set_page_lsn(txn->get_prev_lsn());
            auto &tab_ = sm_manager_->db_.get_table(item->GetTableName());
            for (IndexMeta &index : tab_.indexes)
            {

                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(item->GetTableName(), index.cols)).get();
                char *key = new char[index.col_tot_len];
                int offset = 0;
                for (auto &col : index.cols)
                {
                    memcpy(key + offset, record->data + col.offset, col.len);
                    offset += col.len;
                }
                ih->delete_entry(key, context->txn_);
                delete[] key;
            }

            // DeleteLogRecord* delete_log = new DeleteLogRecord(txn->get_transaction_id(), *record, item->GetRid(), item->GetTableName());
            // delete_log->prev_lsn_ = txn->get_prev_lsn();
            // log_manager->add_log_to_buffer(delete_log);
            // txn->set_prev_lsn(delete_log->lsn_);
            // delete delete_log;
        }
        else if (item->GetWriteType() == WType::DELETE_TUPLE)
        {
            auto record = item->GetRecord();
            auto rid_ = file_handle->insert_record(record.data, context);
            //add in 11
            file_handle->fetch_page_handle(item->GetRid().page_no).page->set_page_lsn(txn->get_prev_lsn());
            auto &tab_ = sm_manager_->db_.get_table(item->GetTableName());
            for (IndexMeta &index : tab_.indexes)
            {
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(item->GetTableName(), index.cols)).get();
                char* key = new char[index.col_tot_len];
                int offset = 0;
                for (auto &col : index.cols)
                {
                    memcpy(key + offset, record.data + col.offset, col.len);
                    offset += col.len;
                }
                ih->insert_entry(key, rid_, context->txn_);
                delete[] key;
            }
            
            InsertLogRecord* insert_log = new InsertLogRecord(txn->get_transaction_id(), item->GetRecord(), item->GetRid(), item->GetTableName());
            insert_log->prev_lsn_ = txn->get_prev_lsn();
            log_manager->add_log_to_buffer(insert_log);
            txn->set_prev_lsn(insert_log->lsn_);
            delete insert_log;
        }
        else if (item->GetWriteType() == WType::UPDATE_TUPLE) {
            auto old_record = file_handle->get_record(item->GetRid(), context);
            auto new_record = item->GetRecord();
            auto &tab_ = sm_manager_->db_.get_table(item->GetTableName());
            file_handle->update_record(item->GetRid(), item->GetRecord().data, context);
            //add in 11
            file_handle->fetch_page_handle(item->GetRid().page_no).page->set_page_lsn(txn->get_prev_lsn());
            for (IndexMeta &index : tab_.indexes)
            {

                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(item->GetTableName(), index.cols)).get();
                char *new_key = new char[index.col_tot_len];
                char *old_key = new char[index.col_tot_len];
                int offset = 0;
                for (auto &col : index.cols)
                {
                    memcpy(new_key + offset, new_record.data + col.offset, col.len);
                    memcpy(old_key + offset, old_record->data + col.offset, col.len);
                    offset += col.len;
                }
                ih->delete_entry(old_key, context->txn_), 
                ih->insert_entry(new_key, item->GetRid(), context->txn_);
                delete[] new_key;
                delete[] old_key;
            }

            // UpdateLogRecord* update_log = new UpdateLogRecord(txn->get_transaction_id(),
            //                 *old_record, item->GetRecord(), item->GetRid(), item->GetTableName());
            // update_log->prev_lsn_ = txn->get_prev_lsn();
            // log_manager->add_log_to_buffer(update_log);
            // txn->set_prev_lsn(update_log->lsn_);
            // delete update_log;
        }
        delete write_set->back();
        write_set->pop_back();
    }
    write_set->clear();
    auto lock_set = txn->get_lock_set();
    for(auto iter = lock_set->begin(); iter != lock_set->end(); iter ++)
        lock_manager_->unlock(txn, *iter);
    // AbortLogRecord* abort_log = new AbortLogRecord(txn->get_transaction_id());
    // abort_log->prev_lsn_ = txn->get_prev_lsn();
    // log_manager->add_log_to_buffer(abort_log);
    // txn->set_prev_lsn(abort_log->lsn_);
    // log_manager->flush_log_to_disk();
    // txn->set_state(TransactionState::ABORTED);
    // delete context;
    // delete abort_log;
}